import gadapt.ga_model.definitions as definitions
from gadapt.adapters.validation.base_options_validator import BaseOptionsValidator


class CommonOptionsValidator(BaseOptionsValidator):
    """
    Common class for options validation
    """

    def __init__(self, options) -> None:
        super().__init__(options)
        self._validation_messages = []

    def _validate_options(self):
        self.success &= self._check_nones_types_and_values()
        self.success &= self._check_decision_variables()
        self.success &= self._check_cost_function()

    def _check_decision_variables(self) -> bool:
        if self.options._decision_variables is None:
            return False
        rslt = True
        for v in self.options._decision_variables:
            if v.min_value is None or v.max_value is None or v.step is None:
                self._add_message(
                    "Min value, max value and step must not be\
                        None! (Variable {var_no})".format(
                        var_no=v.variable_id
                    )
                )
                rslt &= False
            elif (
                not (isinstance(v.min_value, float) or isinstance(v.min_value, int))
                or not (isinstance(v.max_value, float) or isinstance(v.max_value, int))
                or not (isinstance(v.step, float) or isinstance(v.step, int))
            ):
                self._add_message(
                    "Min value, max value and step must be float or int\
                        values! (Variable {var_no})".format(
                        var_no=v.variable_id
                    )
                )
                rslt &= False
            elif v.min_value > v.max_value:
                self._add_message(
                    "Min value must be less or equal than max value!\
                        (Variable {var_no})".format(
                        var_no=v.variable_id
                    )
                )
                rslt &= False
            elif v.step <= 0.0:
                self._add_message(
                    "Step must be positive float value! (Variable {var_no})".format(
                        var_no=v.variable_id
                    )
                )
                rslt &= False
            elif v.min_value + v.step > v.max_value:
                self._add_message(
                    "Invalid step value! (Variable {var_no})".format(
                        var_no=v.variable_id
                    )
                )
                rslt &= False
            if v.decimal_places < 0:
                self._add_message(
                    "Invalid number of decimal places! (Variable {var_no})".format(
                        var_no=v.variable_id
                    )
                )
                rslt &= False
        return rslt

    def _check_cost_function(self):
        if self.options.cost_function is None:
            self._add_message("Cost Function must not be None!")
            return False
        elif not callable(self.options.cost_function):
            self._add_message("Cost Function must be callable!")
            return False
        argsmin = {}
        argsmax = {}
        for v in self.options._decision_variables:
            argsmin[v.variable_id] = v.min_value
            argsmax[v.variable_id] = v.max_value
        try:
            func_result_min = self.options.cost_function(argsmin)
            func_result_max = self.options.cost_function(argsmax)
            if not isinstance(func_result_min, float) or not isinstance(
                func_result_max, float
            ):
                self._add_message("The function must return float value!")
                return False
        except KeyError:
            self._add_message(
                "Inadequate number of parameters for the passed function!"
            )
            return False
        except Exception as ex:
            self._add_message(
                "An exception raised on the function call: {exc}".format(exc=str(ex))
            )
            return False
        return True

    def _check_nones_types_and_values(self) -> bool:
        rslt = True
        population_size = 0
        if self.options.population_size is None:
            self._add_message("Population size must not be None!")
            rslt &= False
        elif not isinstance(self.options.population_size, int):
            self._add_message("Population size type must be int!")
            rslt &= False
        elif self.options.population_size < 4 or self.options.population_size > 65536:
            self._add_message("Population size must be an int between 4 and 65536!")
            rslt &= False
        else:
            population_size = self.options.population_size
        keep_elitism_percentage = self.options.keep_elitism_percentage
        if not (0 <= self.options.keep_elitism_percentage <= 100):
            self._add_message("Elitism percentage value must be between 0 and 100!")
            rslt &= False
        immigration_number = 0
        if self.options.immigration_number is None:
            self._add_message("Immigration Number must not be None!")
            rslt &= False
        elif not isinstance(self.options.immigration_number, int):
            self._add_message("Immigration Number must be type int!")
            rslt &= False
        elif (
            population_size > 0
            and self.options.immigration_number > population_size - 1
        ):
            self._add_message(
                "Immigration Number must not be type same or greater\
                    than population size!"
            )
            rslt &= False
        elif self.options.immigration_number < 0:
            self._add_message("Immigration Number must ne eqals or greather than 0!")
            rslt &= False
        else:
            immigration_number = self.options.immigration_number
        if (
            self.options.number_of_mutation_chromosomes is None
            and self.options.percentage_of_mutation_chromosomes is None
        ):
            self._add_message(
                "Number of Mutation Chromosomes or Percentage of Mutation\
                    Chromosomes must not be None!"
            )
            rslt &= False
        elif not isinstance(
            self.options.number_of_mutation_chromosomes, int
        ) and not isinstance(self.options.percentage_of_mutation_chromosomes, float):
            self._add_message(
                "Number of Mutation Chromosomes must be type int or Percentage\
                    of Mutation Chromosomes must be float!"
            )
            rslt &= False
        else:
            if (
                (self.options.number_of_mutation_chromosomes is not None)
                and (isinstance(self.options.number_of_mutation_chromosomes, int))
                and (self.options.number_of_mutation_chromosomes > 0)
            ):
                if (
                    population_size > 0
                    and self.options.number_of_mutation_chromosomes
                    > round(population_size * (keep_elitism_percentage / 100))
                    - immigration_number
                ):
                    self._add_message(
                        "Invalid number of mutation chromosomes: {0}".format(
                            str(self.options.number_of_mutation_chromosomes)
                        )
                    )
                    rslt &= False
            elif (self.options.percentage_of_mutation_chromosomes is None) or (
                not isinstance(self.options.percentage_of_mutation_chromosomes, float)
            ):
                self._add_message(
                    "Number of Mutation Chromosomes must have int\
                        value >=0 or Percentage of\
                        Mutation Chromosomes must have float (0.0-100.0) value! "
                )
                rslt &= False
            elif (
                self.options.percentage_of_mutation_chromosomes < 0.0
                or self.options.percentage_of_mutation_chromosomes > 100.0
            ):
                self._add_message(
                    "Invalid percentage of mutation chromosomes: {0} and number\
                        of mutation chromosomes: {1}".format(
                        str(self.options.percentage_of_mutation_chromosomes),
                        str(self.options.number_of_mutation_chromosomes),
                    )
                )
                rslt &= False
        if self.options._decision_variables is None:
            self._add_message("Decision variables must not be None!")
            rslt &= False
        num_of_decision_variables = len(self.options._decision_variables)  # type: ignore
        if num_of_decision_variables < 1:
            self._add_message("At least one decision variable must be added!")
            rslt &= False
        if (
            self.options.number_of_mutation_genes is None
            and self.options.percentage_of_mutation_genes is None
        ):
            self._add_message(
                "Number Of Mutation Genes or Percentage Of\
                    Mutation Genes must not be None!"
            )
            rslt &= False
        elif not isinstance(
            self.options.number_of_mutation_genes, int
        ) and not isinstance(self.options.percentage_of_mutation_genes, float):
            self._add_message(
                "Number Of Mutation Genes must have int value >=0 or Percentage of\
                    Mutation Genes must have float (0.0-100.0) value!"
            )
            rslt &= False
        elif (
            (self.options.number_of_mutation_genes is not None)
            and isinstance(self.options.number_of_mutation_genes, int)
            and (self.options.number_of_mutation_genes > 0)
        ):
            if (
                num_of_decision_variables > 0
                and self.options.number_of_mutation_genes > num_of_decision_variables
            ):
                self._add_message(
                    "Invalid number of mutation genes: {0}".format(
                        str(self.options.number_of_mutation_genes)
                    )
                )
                rslt &= False
        elif (self.options.percentage_of_mutation_genes is None) or (
            not isinstance(self.options.percentage_of_mutation_genes, float)
        ):
            self._add_message(
                "Number of Mutation Genes must have int value or Percentage\
                    of Mutation Genes must have float value! "
            )
            rslt &= False
        elif (
            self.options.percentage_of_mutation_genes < 0.0
            or self.options.percentage_of_mutation_genes > 100.0
        ):
            self._add_message(
                "Invalid percentage of mutation genes: {} and number of\
                    mutation genes: {}".format(
                    str(self.options.percentage_of_mutation_genes),
                    str(self.options.number_of_mutation_genes),
                )
            )
            rslt &= False
        if self.options.max_attempt_no is None:
            self._add_message("Max Attempt No must not be None!")
            rslt &= False
        elif not isinstance(self.options.max_attempt_no, int):
            self._add_message("Max Attempt No must be type int!")
            rslt &= False
        elif self.options.max_attempt_no < 1 or self.options.max_attempt_no > 65536:
            self._add_message("Max Attempt No must be type an int between 1 and 65536!")
            rslt &= False
        if (
            self.options.chromosome_mutation == definitions.CROSS_DIVERSITY
            or definitions.CROSS_DIVERSITY in self.options.chromosome_mutation
        ):
            if self.options.cross_diversity_mutation_gene_sampling is None:
                self._add_message(
                    "Cross Diversity Mutation Gene Sampling must not be None!"
                )
                rslt &= False
            elif not isinstance(
                self.options.cross_diversity_mutation_gene_sampling, str
            ):
                self._add_message(
                    "Cross Diversity Mutation Gene Sampling must be type str!"
                )
                rslt &= False
        if self.options.chromosome_mutation is None:
            self._add_message("Chromosome Mutation must not be None!")
            rslt &= False
        elif not isinstance(self.options.chromosome_mutation, str):
            self._add_message("Chromosome Mutation must be type str!")
            rslt &= False
        elif definitions.PARAM_SEPARATOR in self.options.chromosome_mutation:
            mutator_strings = [
                ms.strip()
                for ms in self.options.chromosome_mutation.split(
                    definitions.PARAM_SEPARATOR
                )
            ]
            for mutator_string in mutator_strings:
                if mutator_string not in definitions.CHROMOSOME_MUTATION_STRINGS:
                    self._add_message(
                        "Invalid value of Chromosome Mutation:\
                            {0}. Allowed values: {1}".format(
                            mutator_string,
                            self._get_allowed_values(
                                definitions.CHROMOSOME_MUTATION_STRINGS
                            ),
                        )
                    )
                    rslt &= False
        elif (
            self.options.chromosome_mutation
            not in definitions.CHROMOSOME_MUTATION_STRINGS
        ):
            self._add_message(
                "Invalid value of Chromosome Mutation: {0}. Allowed values: {1}".format(
                    self.options.chromosome_mutation,
                    self._get_allowed_values(definitions.CHROMOSOME_MUTATION_STRINGS),
                )
            )
            rslt &= False
        if self.options.exit_check is None:
            self._add_message("Exit Check must not be None!")
            rslt &= False
        elif not isinstance(self.options.exit_check, str):
            self._add_message("Exit Check must be type str!")
            rslt &= False
        elif self.options.exit_check not in definitions.EXIT_CRITERIA_STRINGS:
            self._add_message(
                "Invalid value of Exit Check: {0}. Allowed values: {1}".format(
                    self.options.exit_check,
                    self._get_allowed_values(definitions.EXIT_CRITERIA_STRINGS),
                )
            )
            rslt &= False
        if self.options.chromosome_mutation == definitions.CROSS_DIVERSITY:
            if self.options.cross_diversity_mutation_gene_sampling is None:
                self._add_message(
                    "Cross Diversity Mutation Gene Selection must not be None!"
                )
                rslt &= False
            elif not isinstance(
                self.options.cross_diversity_mutation_gene_sampling, str
            ):
                self._add_message(
                    "Cross Diversity Mutation Gene Selection must be type str!"
                )
                rslt &= False
            elif not self._validate_selection(
                self.options.cross_diversity_mutation_gene_sampling,
                "Cross Diversity Mutation Gene Selection",
                self.options.number_of_mutation_genes,
                "Group Size for Cross Diversity Mutation Gene Selection must be below\
                    or equal than the number of mutation genes!",
            ):
                rslt &= False
        if self.options.logging is None:
            self._add_message("Logging must not be None!")
            rslt &= False
        elif not isinstance(self.options.logging, bool):
            self._add_message("Logging must be type bool!")
            rslt &= False
        if self.options.must_mutate_for_same_parents is None:
            self._add_message("Must Mutate For Same Parents must not be None!")
            rslt &= False
        elif not isinstance(self.options.must_mutate_for_same_parents, bool):
            self._add_message("Must Mutate For Same Parents must be type bool!")
            rslt &= False
        if (
            self.options.population_mutation == definitions.COST_DIVERSITY
            or definitions.PARENT_DIVERSITY in self.options.population_mutation
        ):
            if self.options.parent_diversity_mutation_chromosome_sampling is None:
                self._add_message(
                    "Parent Diversity Mutation Chromosome Sampling must not be None!"
                )
                rslt &= False
            elif not isinstance(
                self.options.parent_diversity_mutation_chromosome_sampling, str
            ):
                self._add_message(
                    "Parent Diversity Mutation Chromosome Sampling must be type str!"
                )
                rslt &= False
            elif not self._validate_selection(
                self.options.parent_diversity_mutation_chromosome_sampling,
                "Parents Diversity Mutation Chromosome Sampling",
                round(population_size * (keep_elitism_percentage / 100)) - self.options.immigration_number,  # type: ignore
                "Group Size for Parents Diversity Mutation Chromosome\
                Sampling cannot have the value below half population!",
            ):
                rslt &= False
        if self.options.population_mutation is None:
            self._add_message("Population Mutation must not be None!")
            rslt &= False
        elif not isinstance(self.options.population_mutation, str):
            self._add_message("Population Mutation must be type str!")
            rslt &= False
        elif definitions.PARAM_SEPARATOR in self.options.population_mutation:
            mutator_strings = [
                ms.strip()
                for ms in self.options.population_mutation.split(
                    definitions.PARAM_SEPARATOR
                )
            ]
            for mutator_string in mutator_strings:
                if mutator_string not in definitions.POPULATION_MUTATION_STRINGS:
                    self._add_message(
                        "Invalid value of Population Mutation:\
                            {0}. Allowed values: {1}".format(
                            mutator_string,
                            self._get_allowed_values(
                                definitions.POPULATION_MUTATION_STRINGS
                            ),
                        )
                    )
                    rslt &= False
        elif (
            self.options.population_mutation
            not in definitions.POPULATION_MUTATION_STRINGS
        ):
            self._add_message(
                "Invalid value of Population Mutation: {0}. Allowed values: {1}".format(
                    self.options.population_mutation,
                    self._get_allowed_values(definitions.POPULATION_MUTATION_STRINGS),
                )
            )
            rslt &= False
        if self.options.gene_mutation is None:
            self._add_message("Gene Mutation must not be None!")
            rslt &= False
        elif not isinstance(self.options.gene_mutation, str):
            self._add_message("Gene Mutation must be type str!")
            rslt &= False
        elif definitions.PARAM_SEPARATOR in self.options.gene_mutation:
            mutator_strings = [
                ms.strip()
                for ms in self.options.gene_mutation.split(definitions.PARAM_SEPARATOR)
            ]
            for mutator_string in mutator_strings:
                if mutator_string not in definitions.GENE_MUTATION_STRINGS:
                    self._add_message(
                        "Invalid value of Gene Mutation:\
                            {0}. Allowed values: {1}".format(
                            mutator_string,
                            self._get_allowed_values(definitions.GENE_MUTATION_STRINGS),
                        )
                    )
                    rslt &= False
        elif self.options.gene_mutation not in definitions.GENE_MUTATION_STRINGS:
            self._add_message(
                "Invalid value of Gene Mutation: {0}. Allowed values: {1}".format(
                    self.options.gene_mutation,
                    self._get_allowed_values(definitions.GENE_MUTATION_STRINGS),
                )
            )
            rslt &= False
        if self.options.parent_selection is None:
            self._add_message("Parent Selection must not be None!")
            rslt &= False
        elif not isinstance(self.options.parent_selection, str):
            self._add_message("Parent Selection must be type str!")
            rslt &= False
        elif not self._validate_selection(
            self.options.parent_selection,
            "Parent Selection",
            round(population_size * (keep_elitism_percentage / 100)) - self.options.immigration_number,  # type: ignore
            "Group Size for parent selection cannot have\
            the value below half population!",
        ):
            rslt &= False
        if self.options.requested_cost is None:
            self._add_message("Requested Cost must not be None!")
            rslt &= False
        elif not isinstance(self.options.requested_cost, float):
            self._add_message("Requested Cost must be type float!")
            rslt &= False
        if self.options.timeout is None:
            self._add_message("Timeout must not be None!")
            rslt &= False
        elif not isinstance(self.options.timeout, int):
            self._add_message("Timeout must be type int!")
            rslt &= False
        elif self.options.timeout < 1:
            self._add_message(
                "Invalid timeout value: {0}".format(str(self.options.timeout))
            )
            rslt &= False
        return rslt

    def _validate_selection(
        self,
        selection_string,
        selection_type,
        comparing_number,
        group_size_error_message,
    ) -> bool:
        selection_strings = selection_string.split(definitions.PARAM_SEPARATOR)
        if len(selection_strings) > 2 or (
            len(selection_strings) > 1
            and selection_strings[0] != definitions.TOURNAMENT
        ):
            self._add_message(
                "Invalid {0} value: {1}".format(selection_type, selection_string)
            )
            return False
        if (
            selection_strings[0] == definitions.TOURNAMENT
            and len(selection_strings) > 1
        ):
            group_size = selection_strings[1]
            n_group_size = -1
            try:
                n_group_size = int(group_size)
            except Exception:
                self._add_message(
                    "Invalid {0} value: {1}".format(selection_type, selection_string)
                )
                return False
            if n_group_size < 2:
                self._add_message(
                    "Invalid {0} value: {1}. Minimal group size is 2.".format(
                        selection_type, selection_string
                    )
                )
                return False
            if not isinstance(n_group_size, int):
                self._add_message(
                    "Invalid value for {0}: {1}".format(
                        selection_type, selection_string
                    )
                )
                return False
            elif int(n_group_size) > comparing_number:
                self._add_message(group_size_error_message)
                return False
        if selection_strings[0] not in definitions.SELECTION_STRINGS:
            self._add_message(
                "Invalid {0} value: {1}. Allowed values: {2}".format(
                    selection_type,
                    selection_string,
                    self._get_allowed_values(definitions.SELECTION_STRINGS),
                )
            )
            return False
        return True

    def _get_allowed_values(self, s_list):
        rslt = ""
        for s in s_list:
            rslt += s + ", "
        return rslt[:-2]



"""
Chromosome
"""

from typing import List

import gadapt.adapters.string_operation.ga_strings as ga_strings
import gadapt.ga_model.definitions as definitions
from gadapt.ga_model.decision_variable import DecisionVariable
from gadapt.ga_model.gene import Gene
from gadapt.ga_model.ranking_model import RankingModel


class Chromosome(RankingModel):
    def __init__(
        self,
        population_generation: int,
    ):
        """
        Chromosome class.
        Chromosome is a part of the Population. Chromosome consists of Genes.
        Args:
            population_generation: population generation
        """
        super().__init__()
        self._cost_value = definitions.FLOAT_NAN
        self._is_immigrant = False
        self._population_generation = population_generation
        self._chromosome_id = None
        self._mutated_variables_id_list: List[int] = []
        self._first_mutant_generation = 0
        self._first_immigrant_generation = 0
        self._last_mutant_generation = 0
        self._last_immigrant_generation = 0
        self._chromosome_string = None
        self._mother_id = -1
        self._father_id = -1
        self._is_mutated = False
        self._is_immigrant = False
        self._genes: List[Gene] = []

    def __str__(self) -> str:
        return self._get_chromosome_string()

    def __getitem__(self, index) -> Gene:
        return self._genes[index]

    def __next__(self):
        return next(self._genes)

    def __len__(self):
        return len(self._genes)

    def __iter__(self):
        return ChromosomeIterator(self)

    def _get_sorted(self, key: None = None, reverse: bool = False):
        return sorted(self._genes, key=key, reverse=reverse)

    def append(self, g: Gene):
        """
        Appends a new gene into the chromosom
        """
        self._genes.append(g)

    def clear(self):
        """
        Clears all genes from the chromosome
        """
        self._genes.clear()

    def _to_string(self):
        """
        Converts the chromosome to the string
        """
        return ga_strings.chromosome_to_string(self)

    def set_chromosome_string_none(self):
        """
        Sets the chromosome string to None
        """
        self._chromosome_string = None

    def _get_chromosome_string(self):
        if self._chromosome_string is None:
            self._chromosome_string = self._to_string()
        return self._chromosome_string

    @property
    def number_of_mutation_genes(self):
        """
        The number of mutated genes in the chromosome.
        """
        return self._number_of_mutation_genes

    @number_of_mutation_genes.setter
    def number_of_mutation_genes(self, value):
        self._number_of_mutation_genes = value

    @property
    def chromosome_id(self):
        """
        Id of the chromosme
        """
        return self._chromosome_id

    @chromosome_id.setter
    def chromosome_id(self, value):
        self._chromosome_id = value

    @property
    def cost_value(self):
        """
        Calculated cost value  of the chromosome
        """
        return self._cost_value

    @cost_value.setter
    def cost_value(self, value):
        self._cost_value = value

    @property
    def is_mutated(self) -> bool:
        """
        Indicates if the chromosome is mutated
        """
        return self._is_mutated

    @is_mutated.setter
    def is_mutated(self, value: bool):
        self._is_mutated = value

    @property
    def is_immigrant(self) -> bool:
        """
        Indicates if the chromosome is immigrant
        """
        return self._is_immigrant

    @is_immigrant.setter
    def is_immigrant(self, value: bool):
        self._is_immigrant = value

    @property
    def mother_id(self) -> int:
        """
        ID of mother chromosome
        """
        return self._mother_id

    @mother_id.setter
    def mother_id(self, value: int):
        self._mother_id = value

    @property
    def father_id(self) -> int:
        """
        ID of father chromosome
        """
        return self._father_id

    @father_id.setter
    def father_id(self, value: int):
        self._father_id = value

    def add_gene(
        self, gen_var: DecisionVariable, gen_var_value: float = definitions.FLOAT_NAN
    ):
        """
        Adds a gene to the chromosome
        """
        g = Gene(gen_var, gen_var_value)
        self.append(g)

    @property
    def parent_diversity(self) -> float:
        """
        Diversity of parents
        """
        return self._parent_diversity

    @parent_diversity.setter
    def parent_diversity(self, value: float):
        self._parent_diversity = value

    @property
    def population_generation(self) -> int:
        """
        Population generation in which the chromosome appeared
        """
        return self._population_generation

    @population_generation.setter
    def population_generation(self, value: int):
        self._population_generation = value

    @property
    def chromosome_generation(self) -> int:
        """
        Generation of chromosome. It differs from the Population generation.
        It determines how many generations were needed for this chromosome to arise.
        """
        return self._chromosome_generation

    @chromosome_generation.setter
    def chromosome_generation(self, value: int):
        self._chromosome_generation = value

    @property
    def first_mutant_generation(self) -> int:
        """
        Indicates how many generations passed after a first mutation
        """
        return self._first_mutant_generation

    @first_mutant_generation.setter
    def first_mutant_generation(self, value: int):
        self._first_mutant_generation = value

    @property
    def last_mutant_generation(self) -> int:
        """
        Indicates how many generations passed after a last mutation
        """
        return self._last_mutant_generation

    @last_mutant_generation.setter
    def last_mutant_generation(self, value: int):
        self._last_mutant_generation = value

    @property
    def first_immigrant_generation(self) -> int:
        """
        Indicates how many generations passed after a first immigration
        """
        return self._first_immigrant_generation

    @first_immigrant_generation.setter
    def first_immigrant_generation(self, value: int):
        self._first_immigrant_generation = value

    @property
    def last_immigrant_generation(self) -> int:
        """
        Indicates how many generations passed after a last immigration
        """
        return self._last_immigrant_generation

    @last_immigrant_generation.setter
    def last_immigrant_generation(self, value: int):
        self._last_immigrant_generation = value

    @property
    def succ(self) -> bool:
        """
        Indicates if cost function execution succeded
        """
        return self._succ

    @succ.setter
    def succ(self, value: bool):
        self._succ = value

    @property
    def mutated_variables_id_list(self) -> List[int]:
        """
        List of mutated variables
        """
        return self._mutated_variables_id_list


class ChromosomeIterator:
    def __init__(self, chromosome):
        self.chromosome = chromosome
        self.index = 0

    def __iter__(self):
        return self

    def __next__(self):
        if self.index < len(self.chromosome._genes):
            result = self.chromosome._genes[self.index]
            self.index += 1
            return result
        else:
            raise StopIteration



from abc import ABC, abstractmethod
from datetime import datetime
from typing import Optional

from gadapt.adapters.ga_logging.logging_settings import gadapt_log_info
from gadapt.ga_model.population import Population


class BaseExitChecker(ABC):
    """
    Base class for exit check
    Args:
        max_attempt_no (int): Maximal number of attempts with no improvement,
        for the given criteria.

            After this number of attempts with no improvements, the GA exits
    """

    def __init__(self, max_attempt_no: int) -> None:
        self.max_attempt_no = max_attempt_no
        self.attempt_no = 0
        self.population: Optional[Population] = None

    @property
    def attempt_no(self) -> int:
        return self._attempt_no

    @attempt_no.setter
    def attempt_no(self, value: int):
        self._attempt_no = value

    def check(self, population: Population):
        self.population = population
        if self.population is None:
            raise Exception("Population is None!")
        if self.population is None:
            raise Exception("Population is None!")
        time_diff = (datetime.now() - self.population.start_time).total_seconds()
        if time_diff >= self.population.options.timeout:
            self.population.timeout_expired = True
            return True
        if self._is_exit():
            self.attempt_no += 1
        else:
            self.attempt_no = 0
        if self.attempt_no >= self.max_attempt_no:
            gadapt_log_info("function exit.")
            return True
        return False

    @abstractmethod
    def _is_exit(self) -> bool:
        pass



import random
from typing import List

from gadapt.operations.mutation.chromosome_mutation.base_gene_mutation_rate_determinator import (
    BaseGeneMutationRateDeterminator,
)
from gadapt.operations.mutation.chromosome_mutation.base_gene_mutation_selector import (
    BaseGeneMutationSelector,
)
from gadapt.operations.mutation.gene_mutation.base_gene_mutator import BaseGeneMutator


class ComposedGeneMutationSelector(BaseGeneMutationSelector):
    """
    Allows for the composition of multiple mutation selectors and applies them sequentially to the chromosome.
    """

    def __init__(
        self,
        gene_mutation_rate_determinator: BaseGeneMutationRateDeterminator,
        gene_mutator: BaseGeneMutator,
    ) -> None:
        super().__init__(gene_mutation_rate_determinator, gene_mutator)
        self.selectors: List[BaseGeneMutationSelector] = []

    def append(self, selector: BaseGeneMutationSelector):
        """
        Appends mutation selectors to a list of selectors.
        Args:
            selector: An instance of the BaseGeneMutationSelector class that will be added to the list of selectors.
        """
        self.selectors.append(selector)

    def _mutate_chromosome(self):
        if self.chromosome is None:
            raise Exception("Chromosome must not be null")
        if len(self.selectors) == 0:
            raise Exception("at least one mutator must be added")
        if len(self.selectors) > 1:
            random.shuffle(self.selectors)
        nmg = 0
        number_of_mutation_genes = (
            self._gene_mutation_rate_determinator.get_number_of_mutation_genes(
                self.chromosome, self.number_of_mutation_genes
            )
        )
        if number_of_mutation_genes == 0:
            number_of_mutation_genes = 1
        for s in self.selectors:
            if nmg < number_of_mutation_genes:
                s.number_of_mutation_genes = number_of_mutation_genes - nmg
                s.chromosome = self.chromosome
                mg = s._mutate_chromosome()
                nmg += mg
        return nmg



import math
from typing import List

from gadapt.adapters.ga_logging.logging_settings import gadapt_log_info
from gadapt.ga_model.chromosome import Chromosome
from gadapt.operations.cost_finding.base_cost_finder import BaseCostFinder


class ElitismCostFinder(BaseCostFinder):
    """
    Finding costs for a better half of the population
    """

    def _find_costs_for_population(self):
        if self.population is None:
            raise Exception("population must not be null!")
        chromosomes_for_execution: List[Chromosome] = [
            c
            for c in self.population
            if (math.isnan(c.cost_value))
            or (
                c.is_immigrant
                and c.population_generation == self.population.population_generation
            )
        ]
        for c in chromosomes_for_execution:
            self._execute_function(self.population.options.cost_function, c)
        better_chromosomes: List[Chromosome] = self.population.get_sorted(
            key=lambda x: x.cost_value
        )[: self.population.options.keep_number]
        self.population.best_individual = better_chromosomes[0]
        self.population.min_cost = (
            min(better_chromosomes, key=lambda x: x.cost_value)
        ).cost_value
        better_chromosomes_without_immigrants = better_chromosomes[
            : self.population.options.keep_number
            - self.population.options.immigration_number
        ]
        self.population.avg_cost = sum(
            [c.cost_value for c in better_chromosomes_without_immigrants]
        ) / len(better_chromosomes_without_immigrants)
        self._log_population()
        self.population.clear_and_add_chromosomes(better_chromosomes)

    def _log_population(self):
        if self.population.options.logging:
            gadapt_log_info(str(self.population))



from abc import ABC, abstractmethod


class BaseGeneMutationRateDeterminator(ABC):
    def __init__(self) -> None:
        """
        Provides a framework for determining the number of mutation genes in a chromosome.
        """
        super().__init__()
        self.max_number_of_mutation_genes: int = -1

    def get_number_of_mutation_genes(
        self, chromosome, max_number_of_mutation_genes: int
    ):
        if max_number_of_mutation_genes is None:
            raise Exception("max_number_of_mutation_genes must not be None")
        if not chromosome:
            raise Exception("chromosome must not be None")
        self.chromosome = chromosome
        self.max_number_of_mutation_genes = max_number_of_mutation_genes
        return self._get_number_of_mutation_genes()

    @abstractmethod
    def _get_number_of_mutation_genes(self) -> int:
        pass
